if [ -e last_gener.txt ]; then
  echo "NOTHING TO DO"
else
  aevol_create_2b params.in
  aevol_run_2b -e 100 -p 0

  cp checkpoints/checkpoint_000000100/checkpoint_000000100.ae ckpts/ckpt/my_checkpoint_100.ae
  cp checkpoints/checkpoint_000000100/population_000000100.fa ckpts/ckpt/my_population_100.fa
  cp checkpoints/checkpoint_000000100/setup_000000100.txt     ckpts/ckpt/my_setup_100.txt
  ln -sr ckpts/ckpt/my_population_100.fa ckpts/ckpt/ignored.fa

  ln -sr checkpoints/checkpoint_000000000/checkpoint_000000000.ae ckpts/custom_ckpt/foo.ae
  ln -sr checkpoints/checkpoint_000000000/population_000000000.fa ckpts/custom_ckpt/bar.fasta
  ln -sr checkpoints/checkpoint_000000000/setup_000000000.txt		ckpts/custom_ckpt/baz.txt
fi