// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#ifndef AEVOL_RECOMBALIGNMENTSEARCHER_H
#define AEVOL_RECOMBALIGNMENTSEARCHER_H

#include <array>
#include <utility>

#include "Alignment.h"
#include "AlignmentParams.h"
#include "Dna.h"

namespace aevol {

namespace align {

class RecombAlignmentSearcher {
 public:
  static auto find_alignment(std::array<const Dna*, 2> chrsms,
                             JumpingMT& prng,
                             const AlignmentParams& params) -> Alignment;
};

}  // namespace align

}  // namespace aevol

#endif  // AEVOL_RECOMBALIGNMENTSEARCHER_H
